import numpy as np
from pylanczoscpp import *
from .pylanczos_exception import PyLanczosException


def create_suffix_dict(pairs):
    dict = {}

    for keystr, val in pairs:
        dict[np.dtype(keystr)] = val

    return dict


class PyLanczos:
    _dtype_to_suffix = create_suffix_dict(
        [
            ("single", PyLanczosCppFloat),
            ("double", PyLanczosCppDouble),
            ("longdouble", PyLanczosCppLongDouble),
            ("csingle", PyLanczosCppComplexFloat),
            ("cdouble", PyLanczosCppComplexDouble),
            ("clongdouble", PyLanczosCppComplexLongDouble),
        ]
    )

    def __init__(self, matrix, find_maximum, num_eigs):
        """Constructs Lanczos calculation engine.

        Parameters
        ----------
        matrix : Any numpy or scipy matrix type
            The symmetric (Hermitian) matrix to be diagonalized.
            This matrix will never be changed.
        find_maximum : bool
            `True` to calculate the maximum eigenvalue,
            `False` to calculate the minimum one.
        num_eigs : int
            Number of eigenpairs to be calculated.

        Raises
        ------
        PyLanczosException
            If unsupported `dtype` is specified.

        Note
        ----
        A matrix that has the following attributes/methods are
        actually acceptable:

        - `dtype` attribute, which should be `numpy.dtype`
        - `dot` method, which accepts `numpy.ndarray` and
          returns the dot product (matrix-vector multiplication) result
          as `numpy.ndarray`

        Following `dtype` (and their equivalents) are acceptable: `single`, `double`, `longdouble`, `csingle`,
        `cdouble`, and `clongdouble`.
        """

        if matrix.dtype not in PyLanczos._dtype_to_suffix:
            raise PyLanczosException("Unsupported dtype: {}".format(matrix.dtype))

        self._dtype = matrix.dtype
        self._n = matrix.shape[0]
        self._find_maximum = find_maximum
        self._num_eigs = num_eigs
        self._iteration_counts = []

        def mv_mul(v_in, v_out):
            result = matrix.dot(v_in)
            np.copyto(v_out, result)
            # np.dot(self.matrix, v_in, v_out)  # I need something like this...

        self._mv_mul = mv_mul

    @staticmethod
    def create_custom(mv_mul, n, dtype, find_maximum, num_eigs):
        """Constructs Lanczos calculation engine
        with a custom matrix-vector multiplication function.

        Parameters
        ----------
        mv_mul : lambda(numpy.ndarray, numpy.ndarray)
            The matrix-vector multiplication function, which should multiply
            a symmetric (Hermitian) matrix to an input array (the first
            parameter) and store the result into an output array (the second
            parameter). Each arrays will be passed as `numpy.ndarray`.
        n : int
            The dimension of the matrix, i.e. n for n by n matrix.
        dtype : numpy.dtype or str
            The type of the matrix. Following `dtype` (and their equivalents) are acceptable:
            `single`, `double`, `longdouble`, `csingle`, `cdouble`, or `clongdouble`.
        find_maximum : bool
            `True` to calculate the maximum eigenvalue,
            `False` to calculate the minimum one.
        num_eigs : int
            Number of eigenpairs to be calculated.

        Raises
        ------
        PyLanczosException
            If unsupported `dtype` is specified.

        Note
        ----
        Since the matrix-vector multiplication usually spends
        the most part of the Lanczos calculation time, `mv_mul` should be well optimized.
        It is commonly good to define `mv_mul` with binary-level optimized functions,
        such as `numpy.dot`, `numpy.einsum`, etc.
        """

        dummy_matrix = np.array([], dtype=dtype)
        pylanczos_obj = PyLanczos(dummy_matrix, find_maximum, num_eigs)
        pylanczos_obj._n = n
        pylanczos_obj._mv_mul = mv_mul

        return pylanczos_obj

    def run(self):
        """Executes the Lanczos algorithm.

        Returns
        -------
        numpy.ndarray
            Calculated eigenvalues
        numpy.ndarray
            Calculated eigenvectors
        """

        klass = PyLanczos._dtype_to_suffix[self._dtype]
        engine = klass(self._mv_mul, self._n, self._find_maximum, self._num_eigs)

        eigenvalues, eigenvectors, iteration_counts = engine.run()

        self._iteration_counts = iteration_counts

        return eigenvalues, eigenvectors

    @property
    def iteration_counts(self):
        """Returns the Lanczos iteration counts in the last run.

        Returns
        -------
        list[int]
            Iteration counts in the last run. Each element corresponds to
            consecutive iterations performed to calculate eigenvalues.
        """

        return self._iteration_counts


class Exponentiator:
    _dtype_to_suffix = create_suffix_dict(
        [
            ("single", PyExponentiatorCppFloat),
            ("double", PyExponentiatorCppDouble),
            ("longdouble", PyExponentiatorCppLongDouble),
            ("csingle", PyExponentiatorCppComplexFloat),
            ("cdouble", PyExponentiatorCppComplexDouble),
            ("clongdouble", PyExponentiatorCppComplexLongDouble),
        ]
    )

    def __init__(self, matrix):
        """Constructs exponentiation engine.

        Parameters
        ----------
        matrix : Any numpy or scipy matrix type
            The symmetric (Hermitian) matrix to be exponent.
            This matrix will never be changed.

        Raises
        ------
        PyLanczosException
            If unsupported `dtype` is specified.

        Note
        ----
        A matrix that has the following attributes/methods are
        actually acceptable:

        - `dtype` attribute, which should be `numpy.dtype`
        - `dot` method, which accepts `numpy.ndarray` and
          returns the dot product (matrix-vector multiplication) result
          as `numpy.ndarray`

        Following `dtype` (and their equivalents) are acceptable: `single`, `double`, `longdouble`, `csingle`,
        `cdouble`, and `clongdouble`.
        """

        if matrix.dtype not in Exponentiator._dtype_to_suffix:
            raise PyLanczosException("Unsupported dtype: {}".format(matrix.dtype))

        self._dtype = matrix.dtype
        self._n = matrix.shape[0]
        self._iteration_count = 0

        def mv_mul(v_in, v_out):
            result = matrix.dot(v_in)
            np.copyto(v_out, result)
            # np.dot(self.matrix, v_in, v_out)  # I need something like this...

        self._mv_mul = mv_mul

    @staticmethod
    def create_custom(mv_mul, n, dtype):
        """Constructs exponentiation engine
        with a custom matrix-vector multiplication function.

        Parameters
        ----------
        mv_mul : lambda(numpy.ndarray, numpy.ndarray)
            The matrix-vector multiplication function, which should multiply
            a symmetric (Hermitian) matrix to an input array (the first
            parameter) and store the result into an output array (the second
            parameter). Each arrays will be passed as `numpy.ndarray`.
        n : int
            The dimension of the matrix, i.e. n for n by n matrix.
        dtype : numpy.dtype or str
            The type of the matrix. Following `dtype` (and their equivalents) are acceptable:
            `single`, `double`, `longdouble`, `csingle`, `cdouble`, or `clongdouble`.

        Raises
        ------
        PyLanczosException
            If unsupported `dtype` is specified.

        Note
        ----
        Since the matrix-vector multiplication usually spends
        the most part of the Lanczos calculation time, `mv_mul` should be well optimized.
        It is commonly good to define `mv_mul` with binary-level optimized functions,
        such as `numpy.dot`, `numpy.einsum`, etc.
        """

        dummy_matrix = np.array([], dtype=dtype)
        pylanczos_obj = Exponentiator(dummy_matrix)
        pylanczos_obj._n = n
        pylanczos_obj._mv_mul = mv_mul

        return pylanczos_obj

    def run(self, a, input):
        """Executes the Lanczos exponentiation.

        Returns
        -------
        numpy.ndarray
            Calculated exponentiated output `exp(a*matrix)*input`

        Parameters
        ----------
        a : dtype
            Scalar factor of the exponential
        input: numpy.ndarray
            Vector to be multiplied
        """

        klass = Exponentiator._dtype_to_suffix[self._dtype]
        engine = klass(self._mv_mul, self._n)

        output, iteration_count = engine.run(a, input)

        self._iteration_count = iteration_count

        return output

    @property
    def iteration_count(self):
        """Returns the Lanczos iteration count in the last run.

        Returns
        -------
        int
            Iteration count in the last run.
        """

        return self._iteration_count



import numpy as np
import unittest
from pylanczos import PyLanczos
from scipy.sparse import lil_matrix, csr_matrix


class PyLanczosTest(unittest.TestCase):
    def test_simple_matrix(self):
        matrix = np.array([[2, 1, 1],
                           [1, 2, 1],
                           [1, 1, 2]], dtype='float64')
        n = len(matrix)

        engine = PyLanczos(matrix, True, 1)
        eigenvalues, eigenvectors = engine.run()
        eigval = eigenvalues[0]
        eigvec = eigenvectors[:, 0]

        np.testing.assert_almost_equal(eigval, 4.0)
        sign = np.sign(eigvec[0])
        correct_eigvec = sign/np.sqrt(len(matrix))*np.array([1, 1, 1])
        np.testing.assert_allclose(eigvec, correct_eigvec)

    def test_simple_matrix_single_float(self):
        matrix = np.array([[2, 1, 1],
                           [1, 2, 1],
                           [1, 1, 2]], dtype='float32')

        engine = PyLanczos(matrix, True, 1)
        eigenvalues, eigenvectors = engine.run()
        eigval = eigenvalues[0]
        eigvec = eigenvectors[:, 0]

        np.testing.assert_almost_equal(eigval, 4.0, decimal=3)
        sign = np.sign(eigvec[0])
        correct_eigvec = sign/np.sqrt(len(matrix))*np.array([1, 1, 1], dtype='float32')
        np.testing.assert_allclose(eigvec, correct_eigvec, rtol=1e-3)

    def test_simple_matrix_multiple_eigenpairs(self):
        matrix = np.array(
            [[         6, np.sqrt(2), np.sqrt(2)],
             [np.sqrt(2),          1,          5],
             [np.sqrt(2),          5,          1]],
            dtype='float64') / 4
        n = len(matrix)

        engine = PyLanczos(matrix, True, 3)
        eigenvalues, eigenvectors = engine.run()

        correct_eigenvalues = np.array([2, 1, -1], dtype="float64")
        correct_eigenvectors = np.array(
            [[ np.sqrt(2), -np.sqrt(2),  0],
             [          1,          1,   1],
             [          1,          1,  -1]],
            dtype="float64") # Elements are defined so that v[1, :] are positive.

        np.testing.assert_allclose(eigenvalues, correct_eigenvalues)

        ## Adjust sign, adujust norm, and then compare
        for k in range(correct_eigenvectors.shape[1]):
            sign = 1 if eigenvectors[1, k] >= 0 else -1
            norm = np.linalg.norm(correct_eigenvectors[:, k])
            correct_eigenvectors[:, k] *= sign / norm

        np.testing.assert_allclose(eigenvectors, correct_eigenvectors, atol=np.finfo("float64").eps*1e1)

    def test_sparse_matrix_numpy(self):
        n = 10

        matrix = np.zeros((n, n), dtype='float64')
        for i in range(n-1):
            matrix[i, i+1] = -1
            matrix[i+1, i] = -1

        engine = PyLanczos(matrix, False, 1)
        eigenvalues, eigenvectors = engine.run()
        eigval = eigenvalues[0]
        eigvec = eigenvectors[:, 0]

        np.testing.assert_almost_equal(eigval, -2.0*np.cos(np.pi/(n+1)))
        sign = np.sign(eigvec[0])

        correct_eigvec = sign * np.sin((1+np.array(range(n)))*np.pi/(n+1))
        correct_eigvec /= np.linalg.norm(correct_eigvec)

        np.testing.assert_allclose(eigvec, correct_eigvec)

    def test_sparse_matrix_scipy(self):
        n = 10

        matrix = lil_matrix((n, n), dtype='float64')

        for i in range(n-1):
            matrix[(i, i+1)] = -1
            matrix[(i+1, i)] = -1

        engine = PyLanczos(csr_matrix(matrix), False, 1)
        eigenvalues, eigenvectors = engine.run()
        eigval = eigenvalues[0]
        eigvec = eigenvectors[:, 0]

        np.testing.assert_almost_equal(eigval, -2.0*np.cos(np.pi/(n+1)))
        sign = np.sign(eigvec[0])

        correct_eigvec = sign * np.sin((1+np.array(range(n)))*np.pi/(n+1))
        correct_eigvec /= np.linalg.norm(correct_eigvec)

        np.testing.assert_allclose(eigvec, correct_eigvec)

    def test_sparse_matrix_dynamic(self):
        n = 10

        def mv_mul(v_in, v_out):
            for i in range(n-1):
                v_out[i] += -1.0*v_in[i+1]
                v_out[i+1] += -1.0*v_in[i]

        engine = PyLanczos.create_custom(mv_mul, n, 'float64', False, 1)
        eigenvalues, eigenvectors = engine.run()
        eigval = eigenvalues[0]
        eigvec = eigenvectors[:, 0]

        np.testing.assert_almost_equal(eigval, -2.0*np.cos(np.pi/(n+1)))
        sign = np.sign(eigvec[0])

        correct_eigvec = sign * np.sin((1+np.array(range(n)))*np.pi/(n+1))
        correct_eigvec /= np.linalg.norm(correct_eigvec)

        np.testing.assert_allclose(eigvec, correct_eigvec)


    def test_complex_matrix(self):
        matrix = np.array([[  0, 1j,  1],
                           [-1j,  0, 1j],
                           [  1, -1j, 0]], dtype='complex128')

        engine = PyLanczos(matrix, False, 1)
        eigenvalues, eigenvectors = engine.run()
        eigval = eigenvalues[0]
        eigvec = eigenvectors[:, 0]

        np.testing.assert_almost_equal(eigval, -2.0)
        phase = np.angle(eigvec[0])
        correct_eigvec = np.array([1, 1j, -1])
        correct_eigvec = np.exp(1j*phase)/np.linalg.norm(correct_eigvec)*correct_eigvec
        np.testing.assert_allclose(eigvec, correct_eigvec)

    def test_unsupported_dtype(self):
        matrix = np.array([[2, 1, 1],
                           [1, 2, 1],
                           [1, 1, 2]], dtype='int64')

        with self.assertRaises(Exception):
            PyLanczos(matrix, True)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import unittest
from pylanczos import Exponentiator
from scipy.sparse import lil_matrix, csr_matrix


class ExponentiatorTest(unittest.TestCase):
    def test_exponentiate_real(self):
        matrix = np.array([[2, 1, 1], [1, 2, 1], [1, 1, 2]], dtype="float64")
        input = np.array([1, 0, 0], dtype="float64")
        a = 3
        engine = Exponentiator(matrix)
        output = engine.run(a, input)

        w, v = np.linalg.eig(matrix)
        expected_output = v @ np.diag(np.exp(a * w)) @ v.T @ input
        overlap = np.abs(
            np.dot(output, expected_output)
            / np.linalg.norm(output)
            / np.linalg.norm(expected_output)
        )

        np.testing.assert_almost_equal(overlap, 1.0)

    def test_exponentiate_large_matrix(self):
        t = -1
        n = 100
        matrix = lil_matrix((n, n), dtype="complex128")

        for i in range(n - 1):
            matrix[i, i + 1] = t
            matrix[i + 1, i] = t

        matrix[-1, 0] = t
        matrix[0, -1] = t

        input = np.zeros(n, dtype="complex128")
        input[0] = 1 + 2j
        input[n - 1] = 1 + 2j
        input[n // 2] = 8 + 2j
        input = input / np.linalg.norm(input)  # normalize

        a = 3j

        engine = Exponentiator(csr_matrix(matrix))
        output = engine.run(a, input)

        w, v = ExponentiatorTest.make_plane_wave(t, n)
        expected_output = v @ np.diag(np.exp(a * w)) @ v.conj().T @ input

        overlap = np.abs(
            np.dot(output.conj(), expected_output)
            / np.linalg.norm(output)
            / np.linalg.norm(expected_output)
        )

        np.testing.assert_almost_equal(overlap, 1.0)

    def test_exponentiate_zero_delta(self):
        t = -1
        n = 100
        matrix = lil_matrix((n, n), dtype="complex128")

        for i in range(n - 1):
            matrix[i, i + 1] = t
            matrix[i + 1, i] = t

        matrix[-1, 0] = t
        matrix[0, -1] = t

        input = np.zeros(n, dtype="complex128")
        input[0] = 1 + 2j
        input[n - 1] = 1 + 2j
        input[n // 2] = 8 + 2j
        input = input / np.linalg.norm(input)  # normalize

        a = 0 + 0j

        engine = Exponentiator(csr_matrix(matrix))
        output = engine.run(a, input)

        expected_output = input

        overlap = np.abs(
            np.dot(output.conj(), expected_output)
            / np.linalg.norm(output)
            / np.linalg.norm(expected_output)
        )

        np.testing.assert_almost_equal(overlap, 1.0)

    def test_exponentiate_large_matrix_dynamic(self):
        t = -1
        n = 100

        def mv_mul(v_in, v_out):
            for i in range(n - 1):
                v_out[i] += t * v_in[i + 1]
                v_out[i + 1] += t * v_in[i]

            v_out[-1] += t * v_in[0]
            v_out[0] += t * v_in[-1]

        input = np.zeros(n, dtype="complex128")
        input[0] = 1 + 2j
        input[n - 1] = 1 + 2j
        input[n // 2] = 8 + 2j
        input = input / np.linalg.norm(input)  # normalize

        a = 3j

        engine = Exponentiator.create_custom(mv_mul, n, dtype="complex128")
        output = engine.run(a, input)

        w, v = ExponentiatorTest.make_plane_wave(t, n)
        expected_output = v @ np.diag(np.exp(a * w)) @ v.conj().T @ input

        overlap = np.abs(
            np.dot(output.conj(), expected_output)
            / np.linalg.norm(output)
            / np.linalg.norm(expected_output)
        )

        np.testing.assert_almost_equal(overlap, 1.0)

    def make_plane_wave(t, n):
        ks = 2 * np.pi / n * np.arange(n)
        es = 2 * t * np.cos(ks)

        u = np.ndarray((n, n), dtype="complex128")

        for j, k in enumerate(ks):
            u[:, j] = np.exp(1j * k * np.arange(n)) / np.sqrt(n)

        return (es, u)


if __name__ == "__main__":
    unittest.main()
